from torchmeta.transforms.categorical import Categorical, FixedCategory
from torchmeta.transforms.augmentations import Rotation, HorizontalFlip, VerticalFlip
from torchmeta.transforms.splitters import Splitter, ClassSplitter, WeightedClassSplitter
from torchmeta.transforms.target_transforms import TargetTransform, DefaultTargetTransform, SegmentationPairTransform
